@@ -29,12 +29,80 @@
 #include "paragraph.h"
 #include "utils.h"
 #include "text.h"
+#include <glib/gi18n.h>
 
 static void gtk_html_a11y_class_init (GtkHTMLA11YClass *klass);
 static void gtk_html_a11y_init       (GtkHTMLA11Y *a11y);
 
 static GtkAccessibleClass *parent_class = NULL;
 
+
+static gint
+get_n_actions (AtkAction *action)
+{
+	return 1;
+}
+
+static G_CONST_RETURN gchar*
+get_description (AtkAction *action, gint i)
+{
+	if (i == 0)
+		return _("grab focus");
+
+	return NULL;
+}
+
+static G_CONST_RETURN gchar*
+action_get_name (AtkAction *action, gint i)
+{
+	if (i == 0)
+		return _("grab focus");
+
+	return NULL;
+}
+
+
+
+static gboolean
+do_action (AtkAction * action, gint i)
+{
+	GtkWidget *widget;
+	gboolean return_value = TRUE;
+
+	widget = GTK_ACCESSIBLE (action)->widget;
+
+	if (widget == NULL) {
+	/*
+	* State is defunct
+	*/
+	return FALSE;
+	}
+
+	if (!GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+		return FALSE;
+
+
+	switch (i) {
+	case 0:
+		gtk_widget_grab_focus (widget);
+	default:
+		return_value = FALSE;
+		break;
+	}
+	return return_value;
+}
+
+static void
+atk_action_interface_init (AtkActionIface *iface)
+{
+	g_return_if_fail (iface != NULL);
+
+	iface->do_action = do_action;
+	iface->get_n_actions = get_n_actions;
+	iface->get_description = get_description;
+	iface->get_name = action_get_name;
+}
+
 GType
 gtk_html_a11y_get_type (void)
 {
@@ -54,23 +122,29 @@ gtk_html_a11y_get_type (void)
 			NULL                                             /* value table */
 		};
 
+		static const GInterfaceInfo atk_action_info = {
+			(GInterfaceInitFunc) atk_action_interface_init,
+			(GInterfaceFinalizeFunc) NULL,
+			NULL
+		};
+
 		/*
 		 * Figure out the size of the class and instance 
 		 * we are deriving from
 		 */
 		AtkObjectFactory *factory;
-		GType derived_type;
 		GTypeQuery query;
 		GType derived_atk_type;
 
-		derived_type = g_type_parent (GTK_TYPE_HTML);
-		factory = atk_registry_get_factory (atk_get_default_registry (), derived_type);
+		factory = atk_registry_get_factory (atk_get_default_registry (), GTK_TYPE_WIDGET);
 		derived_atk_type = atk_object_factory_get_accessible_type (factory);
 		g_type_query (derived_atk_type, &query);
 		tinfo.class_size = query.class_size;
 		tinfo.instance_size = query.instance_size;
 
 		type = g_type_register_static (derived_atk_type, "GtkHTMLA11Y", &tinfo, 0);
+
+		g_type_add_interface_static (type, ATK_TYPE_ACTION, &atk_action_info);
 	}
 
 	return type;
@@ -127,6 +201,16 @@ gtk_html_a11y_ref_child (AtkObject *accessible, gint index)
 	return accessible_child;
 }
 
+static G_CONST_RETURN gchar*
+gtk_html_a11y_get_name (AtkObject *obj)
+{
+	G_CONST_RETURN gchar *name;
+	if (obj->name != NULL) {
+		return obj->name;
+	}
+	return _("Panel containing HTML");
+}
+
 static void
 gtk_html_a11y_class_init (GtkHTMLA11YClass *klass)
 {
@@ -138,6 +222,7 @@ gtk_html_a11y_class_init (GtkHTMLA11YClass *klass)
 	atk_class->initialize = gtk_html_a11y_initialize;
 	atk_class->get_n_children = gtk_html_a11y_get_n_children;
 	atk_class->ref_child = gtk_html_a11y_ref_child;
+	atk_class->get_name = gtk_html_a11y_get_name;
 
 	gobject_class->finalize = gtk_html_a11y_finalize;
 }
@@ -180,17 +265,19 @@ gtk_html_a11y_grab_focus_cb(GtkWidget * widget)
 
 }
 
+
+static AtkObject * gtk_html_a11y_focus_object = NULL;
+
 static void
-gtk_html_a11y_cursor_move_cb(GtkWidget *widget,  GtkDirectionType dir_type, GtkHTMLCursorSkipType skip)
+gtk_html_a11y_cursor_changed_cb (GtkWidget *widget)
 {
         AtkObject *focus_object, *obj;
-	static AtkObject * prev_object = NULL;
 
 	focus_object = gtk_html_a11y_get_focus_object (widget);
        obj = gtk_widget_get_accessible (widget);
 	
-	if (prev_object != focus_object) {
-		prev_object = focus_object;
+	if (gtk_html_a11y_focus_object != focus_object) {
+		gtk_html_a11y_focus_object = focus_object;
         	g_object_set_data (G_OBJECT(obj), "gail-focus-object", focus_object);
         	atk_focus_tracker_notify (focus_object);
 	} else {
@@ -203,6 +290,47 @@ gtk_html_a11y_cursor_move_cb(GtkWidget *widget,  GtkDirectionType dir_type, GtkH
         }
 }
 
+static void
+gtk_html_a11y_insert_object_cb (GtkWidget * widget, int pos, int len) 
+{
+	AtkObject * a11y, *obj;
+	HTMLText * text;
+
+        obj = gtk_widget_get_accessible (widget);
+	a11y = gtk_html_a11y_get_focus_object (widget);
+
+	if (gtk_html_a11y_focus_object != a11y) {
+		gtk_html_a11y_focus_object = a11y;
+        	g_object_set_data (G_OBJECT(obj), "gail-focus-object", a11y);
+        	atk_focus_tracker_notify (a11y);
+	}
+
+	if (G_IS_HTML_A11Y_TEXT(a11y)) {
+		g_signal_emit_by_name (a11y, "text_changed::insert", pos, len);
+
+	} 
+}
+
+static void
+gtk_html_a11y_delete_object_cb (GtkWidget * widget, int pos, int len) 
+{
+	AtkObject * a11y, *obj;
+	HTMLText * text;
+
+        obj = gtk_widget_get_accessible (widget);
+	a11y = gtk_html_a11y_get_focus_object (widget);
+
+	if (gtk_html_a11y_focus_object != a11y) {
+		gtk_html_a11y_focus_object = a11y;
+        	g_object_set_data (G_OBJECT(obj), "gail-focus-object", a11y);
+        	atk_focus_tracker_notify (a11y);
+	}
+
+	if (G_IS_HTML_A11Y_TEXT(a11y)) {
+		g_signal_emit_by_name (a11y, "text_changed::delete", pos, len);
+	}
+}
+
 AtkObject* 
 gtk_html_a11y_new (GtkWidget *widget)
 {
@@ -216,13 +344,20 @@ gtk_html_a11y_new (GtkWidget *widget)
 	accessible = ATK_OBJECT (object);
 	atk_object_initialize (accessible, widget);
 
-	accessible->role = ATK_ROLE_HTML_CONTAINER;
-	g_signal_connect_after(widget, "grab_focus", 
+	accessible->role = ATK_ROLE_PANEL;
+	g_signal_connect_after (widget, "grab_focus", 
 			G_CALLBACK (gtk_html_a11y_grab_focus_cb),
 			NULL);
-	g_signal_connect_after(widget, "cursor_move",
-			G_CALLBACK(gtk_html_a11y_cursor_move_cb),
+	g_signal_connect (widget, "cursor_changed",
+			G_CALLBACK(gtk_html_a11y_cursor_changed_cb),
 			NULL);
+	g_signal_connect_after (widget, "object_inserted",
+			G_CALLBACK(gtk_html_a11y_insert_object_cb),
+			NULL);
+	g_signal_connect_after (widget, "object_deleted",
+			G_CALLBACK(gtk_html_a11y_delete_object_cb),
+			NULL);
+
 	html_utils_get_accessible(GTK_HTML(widget)->engine->clue, accessible);
 
 	/* printf ("created new gtkhtml accessible object\n"); */
