@@ -22,7 +22,7 @@
 */
 
 #include <config.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include "control-data.h"
 #include "spell.h"
 #include "properties.h"
@@ -47,6 +47,9 @@ gtk_html_control_data_new (GtkHTML *html, GtkWidget *vbox)
 	ncd->search_text             = NULL;
 	ncd->replace_text_search     = NULL;
 	ncd->replace_text_replace    = NULL;
+	ncd->has_spell_control_set   = FALSE;
+	ncd->language                = NULL;
+	ncd->icon_theme              = gnome_icon_theme_new ();
 
 	spell_init (html, ncd);
 
@@ -79,5 +82,10 @@ gtk_html_control_data_destroy (GtkHTMLControlData *cd)
 	if (cd->languages)
 		CORBA_free (cd->languages);
 
+	if (cd->icon_theme) {
+		g_object_unref (cd->icon_theme);
+		cd->icon_theme = NULL;
+	}
+
 	g_free (cd);
 }
