@@ -22,7 +22,7 @@
 */
 
 #include <config.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -539,16 +539,8 @@ font_style_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname
                                                                               
         /* g_warning ("wowee %s :: %s", path, state); */
         for (i = 0; font_style_assoc[i].verb != NULL; i++) {
-                if (!strcmp (cname, font_style_assoc[i].verb)) {
-                        if (font_style_assoc[i].style > GTK_HTML_FONT_STYLE_MAX)
-                                gtk_html_set_font_style (cd->html, ~0,
-                                                         font_style_assoc[i].style);
-                        else
-                                gtk_html_set_font_style (cd->html,
-                                                         GTK_HTML_FONT_STYLE_MAX
-                                                         & ~GTK_HTML_FONT_STYLE_SIZE_MASK,
-                                                         font_style_assoc[i].style);
-                }
+                if (!strcmp (cname, font_style_assoc[i].verb))
+			gtk_html_toggle_font_style (cd->html, font_style_assoc[i].style);
         }
 }
 
@@ -730,6 +722,9 @@ static EditorUIPixmap pixmaps_map [] =
 	{ "/Toolbar/EditPaste", "stock_paste", EDITOR_ICON_TOOLBAR },
 	{ "/Toolbar/EditFind", "stock_search", EDITOR_ICON_TOOLBAR },
 	{ "/Toolbar/InsertImage", "stock_insert_image", EDITOR_ICON_TOOLBAR },
+	{ "/Toolbar/InsertLink", "stock_insert-url", EDITOR_ICON_TOOLBAR },
+	{ "/Toolbar/InsertRule", "stock_insert-rule", EDITOR_ICON_TOOLBAR },
+	{ "/Toolbar/InsertTable", "stock_insert-table", EDITOR_ICON_TOOLBAR },
 
 	{ "/menu/Edit/EditUndoRedo/EditUndo", "stock_undo", EDITOR_ICON_MENU },
 	{ "/menu/Edit/EditUndoRedo/EditRedo", "stock_redo", EDITOR_ICON_MENU },
@@ -744,6 +739,9 @@ static EditorUIPixmap pixmaps_map [] =
 	{ "/menu/Edit/EditMisc/EditSpellCheck", "stock_spellcheck", EDITOR_ICON_MENU },
 
 	{ "/menu/Insert/Component/InsertImage", "stock_insert_image", EDITOR_ICON_MENU },
+	{ "/menu/Insert/Component/InsertLink", "stock_insert-url", EDITOR_ICON_MENU },
+	{ "/menu/Insert/Component/InsertRule", "stock_insert-rule", EDITOR_ICON_MENU },
+	{ "/menu/Insert/Component/InsertTable", "stock_insert-table", EDITOR_ICON_MENU },
 };
 
 void
@@ -776,10 +774,21 @@ menubar_setup (BonoboUIComponent  *uic,
 
 	for (i = 0; i < sizeof (pixmaps_map) / sizeof (pixmaps_map [0]); i ++)
 	{
-		bonobo_ui_component_set_prop (uic, pixmaps_map [i].path, "pixtype", "filename", NULL);
-		bonobo_ui_component_set_prop (uic, pixmaps_map [i].path, "pixname",
-					      gnome_icon_theme_lookup_icon (cd->icon_theme, pixmaps_map [i].stock_name, pixmaps_map [i].size, NULL, NULL),
-					      NULL);
+		char *filename;
+
+		filename = gnome_icon_theme_lookup_icon (cd->icon_theme, pixmaps_map [i].stock_name, pixmaps_map [i].size, NULL, NULL);
+
+		if (filename) {
+			GdkPixbuf *pixbuf;
+			char *xml_pixbuf;
+
+			pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
+			xml_pixbuf = bonobo_ui_util_pixbuf_to_xml (pixbuf);
+			g_object_unref (pixbuf);
+
+			bonobo_ui_component_set_prop (uic, pixmaps_map [i].path, "pixname", xml_pixbuf, NULL);
+		} else
+			g_warning ("cannot find icon: '%s' in gnome icon theme", pixmaps_map [i].stock_name);
 	}
 
 	spell_create_language_menu (cd);
