@@ -373,17 +373,26 @@ draw (HTMLObject *o,
 {
 	HTMLObject *aclue;
 	HTMLClueV *cluev;
+	GdkRectangle paint;
 
 	cluev = HTML_CLUEV (o);
+	
+	if (!html_object_intersect (o, &paint, x, y, width, height))
+		return;
+
+	if (cluev->background_color) {
+		html_painter_alloc_color (p, &cluev->background_color->color);
+		html_painter_draw_background (p, 
+					      &cluev->background_color->color,
+					      NULL, tx + paint.x, ty + paint.y, paint.width, paint.height, 0, 0);
+	}
+
 	HTML_OBJECT_CLASS (&html_clue_class)->draw (o,
 						    p,
 						    x, y ,
 						    width, height,
 						    tx, ty);
 
-	if (y + height < o->y - o->ascent || y > o->y + o->descent)
-		return;
-
 	tx += o->x;
 	ty += o->y - o->ascent;
 
@@ -865,6 +874,16 @@ get_right_clear (HTMLClue *self,
 	return y;
 }
 
+static HTMLDirection
+html_cluev_real_get_direction (HTMLObject *o)
+{
+	if (HTML_CLUEV (o)->dir == HTML_DIRECTION_DERIVED && o->parent) {
+		return html_object_get_direction (o->parent);
+	}
+
+	return HTML_CLUEV (o)->dir;
+}
+
 
 void
 html_cluev_type_init (void)
@@ -897,6 +916,7 @@ html_cluev_class_init (HTMLClueVClass *klass,
 	object_class->check_point = check_point;
 	object_class->get_left_margin = get_left_margin;
 	object_class->get_right_margin = get_right_margin;
+	object_class->get_direction = html_cluev_real_get_direction;
 
 	clue_class->get_left_clear = get_left_clear;
 	clue_class->get_right_clear = get_right_clear;
@@ -927,7 +947,8 @@ html_cluev_init (HTMLClueV *cluev,
 	object->percent = percent;
 
 	clue->valign = HTML_VALIGN_BOTTOM;
-	clue->halign = HTML_HALIGN_LEFT;
+	clue->halign = HTML_HALIGN_NONE;
+	cluev->dir = HTML_DIRECTION_DERIVED;
 	clue->head = clue->tail = clue->curr = 0;
 
 	cluev->align_left_list = 0;
@@ -936,6 +957,7 @@ html_cluev_init (HTMLClueV *cluev,
 	cluev->border_style = HTML_BORDER_NONE;
 	cluev->border_width = 0;
 	cluev->border_color = NULL;
+	cluev->background_color = NULL;
 }
 
 HTMLObject *
@@ -950,23 +972,33 @@ html_cluev_new (gint x, gint y, gint percent)
 }
 
 void
-html_cluev_set_border (HTMLClueV *cluev, HTMLStyle *style)
+html_cluev_set_style (HTMLClueV *cluev, HTMLStyle *style)
 {
 	if (style != NULL) {
 		if (cluev->border_color)
 			html_color_unref (cluev->border_color);
 
+		cluev->padding = style->padding;
+
 		cluev->border_style = style->border_style;
 		cluev->border_width = style->border_width;
 		cluev->border_color = style->border_color;
 		if (cluev->border_color)
 			html_color_ref (cluev->border_color);
+
+		cluev->background_color = style->bg_color;
+		if (cluev->background_color)
+			html_color_ref (cluev->background_color);
 	} else {
 		if (cluev->border_color)
 			html_color_unref (cluev->border_color);
 
+		if (cluev->border_color)
+			html_color_unref (cluev->background_color);
+
 		cluev->border_style = HTML_BORDER_NONE;
 		cluev->border_width = 0;
 		cluev->border_color = NULL;
+		cluev->background_color = NULL;
 	}
 }
