@@ -57,7 +57,6 @@ static void         spell_error_destroy     (SpellError *se);
 static void         move_spell_errors       (GList *spell_errors, guint offset, gint delta);
 static GList *      remove_spell_errors     (GList *spell_errors, guint offset, guint len);
 static void         remove_text_slaves      (HTMLObject *self);
-static void         html_text_change_attrs  (PangoAttrList *attr_list, GtkHTMLFontStyle style, HTMLEngine *e, gint start_index, gint end_index, gboolean avoid_default_size);
 
 /* void
 debug_spell_errors (GList *se)
@@ -99,6 +98,7 @@ html_text_pango_info_destroy (HTMLTextPangoInfo *pi)
 		pango_item_free (pi->entries [i].item);
 		g_free (pi->entries [i].widths);
 	}
+	g_free (pi->entries);
 	g_free (pi->attrs);
 	g_free (pi->face);
 	g_free (pi);
@@ -964,7 +964,7 @@ html_text_calc_part_width (HTMLText *text, HTMLPainter *painter, char *start, gi
 		} else {
 			width += pi->entries [idx].widths [offset];
 			line_offset ++;
-		}
+  		}
 		len --;
 		if (offset >= pi->entries [idx].item->num_chars - 1) {
 			idx ++;
@@ -975,9 +975,9 @@ html_text_calc_part_width (HTMLText *text, HTMLPainter *painter, char *start, gi
 		} else
 			offset ++;
 		s = g_utf8_next_char (s);
-	}
-
- out:
+  	}
+  
+out:
 	if (asc)
 		*asc = html_painter_pango_to_engine (painter, ascent);
 	if (dsc)
@@ -1145,10 +1145,19 @@ html_text_remove_unwanted_line_breaks (char *s, int len, PangoLogAttr *attrs)
 
 		if (attrs [i].is_line_break) {
 			if (last_uc == '.' || last_uc == '/' ||
-			    last_uc == '(' || last_uc == ')' ||
-			    last_uc == '{' || last_uc == '}' ||
-			    last_uc == '[' || last_uc == ']' ||
-			    last_uc == '<' || last_uc == '>')
+			    last_uc == '-' || last_uc == '$' ||
+			    last_uc == '+' || last_uc == '?' ||
+			    last_uc == ')' ||
+			    last_uc == '}' ||
+			    last_uc == ']' ||
+			    last_uc == '>')
+				attrs [i].is_line_break = 0;
+			else if ((uc == '(' ||
+				  uc == '{' ||
+				  uc == '[' ||
+				  uc == '<'
+				  )
+				 && i > 0 && !attrs [i - 1].is_white)
 				attrs [i].is_line_break = 0;
 		}
 		s = g_utf8_next_char (s);
@@ -1306,6 +1315,7 @@ html_text_get_pango_info (HTMLText *text, HTMLPainter *painter)
 			pango_glyph_string_get_logical_widths (glyphs, translated + item->offset, item->length, item->analysis.level, text->pi->entries [i].widths);
 			pango_glyph_string_free (glyphs);
 		}
+
 		g_free (heap);
 		g_list_free (items);
 	}
@@ -2448,9 +2458,11 @@ html_text_init (HTMLText *text,
 
 	html_object_init (HTML_OBJECT (text), HTML_OBJECT_CLASS (klass));
 
-	text->text_bytes    = html_text_sanitize (&str, &len);
-	text->text_len      = len;
-	text->text          = g_memdup (str, text->text_bytes + 1);
+	text->text_bytes = html_text_sanitize (&str, &len);
+	text->text_len = len;
+	text->text = g_memdup (str, text->text_bytes + 1);
+	text->text [text->text_bytes] = '\0';
+
 	text->font_style    = font_style;
 	text->face          = NULL;
 	text->color         = color;
@@ -2530,7 +2542,8 @@ html_text_set_text (HTMLText *text, const gchar *new_text)
 	g_free (text->text);
 	text->text_len = -1;
 	text->text_bytes = html_text_sanitize (&new_text, &text->text_len);
-	text->text = g_memdup (new_text, text->text_bytes);
+	text->text = g_memdup (new_text, text->text_bytes + 1);
+	text->text [text->text_bytes] = '\0';
 	html_object_change_set (HTML_OBJECT (text), HTML_CHANGE_ALL);
 }
 
@@ -3197,6 +3210,9 @@ style_from_attrs (PangoAttrIterator *iter)
 		case PANGO_ATTR_SIZE:
 			style |= ((HTMLPangoAttrFontSize *) attr)->style;
 			break;
+		case PANGO_ATTR_FAMILY:
+			style |= GTK_HTML_FONT_STYLE_FIXED;
+			break;
 		default:
 			break;
 		}
@@ -3253,7 +3269,7 @@ html_text_get_style_conflicts (HTMLText *text, GtkHTMLFontStyle style, gint star
 	return conflicts;
 }
 
-static void
+void
 html_text_change_attrs (PangoAttrList *attr_list, GtkHTMLFontStyle style, HTMLEngine *e, gint start_index, gint end_index, gboolean avoid_default_size)
 {
 	PangoAttribute *attr;
@@ -3345,6 +3361,10 @@ unset_style_filter (PangoAttribute *attr, gpointer data)
 		if (((HTMLPangoAttrFontSize *) attr)->style & style)
 			return TRUE;
 		break;
+	case PANGO_ATTR_FAMILY:
+		if (style & GTK_HTML_FONT_STYLE_FIXED)
+			return TRUE;
+		break;
 	default:
 		break;
 	}
